
#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "tree.h"
#include "tokens.h"
#include "asdf-tree.h"
#include "ggc.h"

tree
build_nl_string( const char *str )
{
  tree ret;
  int len = strlen( str );
  int size = sizeof(struct nl_string) - 1 + (len+1);

  ret = ggc_alloc_tree( size );
  memset( ret, 0, sizeof(struct tree_common) );
  
  NL_STRING_LENGTH(ret) = len;
  strcpy( NL_STRING_POINTER(ret), str );

  TREE_SET_CODE( ret, NL_STRING );

  return ret;
}

tree 
copy_nl_string( tree str )
{
  int size = NL_STRING_SIZE(str);
  tree ret = ggc_alloc_tree( size );

  memmove( ret, str, size );

  TREE_CHAIN(ret) = 0;

  return ret;
}


enum asdf_tree_node_structure_enum 
asdf_tree_node_structure(union lang_tree_node *t)
{
  switch (TREE_CODE (&t->generic)) 
    {
    case NL_DEF_STMT:		return TS_NL_DEF_STMT;
    case IDENTIFIER_NODE:	return TS_ASDF_IDENTIFIER;
    case ASDF_BLOCK:		return TS_ASDF_BLOCK;
    case NL_LOCAL_VAR_EXPR:	return TS_NL_LOCAL_VAR_EXPR;
    case NL_VAR_STMT:		return TS_NL_VAR_STMT;
    case NL_SPECIAL_VALUE:	return TS_NL_SPECIAL_VALUE;
    case NL_NATIVE_FUNCTION:	return TS_NL_NATIVE_FUNCTION;
    case NL_NAMESPACE:		return TS_NL_NAMESPACE;
    case NL_EXPR_STMT:		return TS_NL_EXPR_STMT;
    case NL_NAMESPACE_BLOCK:	return TS_NL_NAMESPACE_BLOCK;
    case NL_ENTRY:		return TS_NL_ENTRY;
    case NL_CONS:		return TS_NL_CONS;
    case NL_STRING:		return TS_NL_STRING;
    case NL_FUNCTION:		return TS_NL_FUNCTION;
    default:			return TS_ASDF_GENERIC;
    }
}

size_t
asdf_tree_size (enum tree_code code)
{
  switch (code)
    {
    case NL_DEF_STMT:		return sizeof (struct nl_def_stmt);
    case ASDF_BLOCK:		return sizeof (struct asdf_block);
    case NL_LOCAL_VAR_EXPR:	return sizeof (struct nl_local_var_expr);
    case NL_VAR_STMT:		return sizeof (struct nl_var_stmt);
    case NL_SPECIAL_VALUE:	return sizeof (struct nl_special_value);
    case NL_NATIVE_FUNCTION:	return sizeof (struct nl_native_function);
    case NL_NAMESPACE:		return sizeof (struct nl_namespace);
    case NL_EXPR_STMT:		return sizeof (struct nl_expr_stmt);
    case NL_NAMESPACE_BLOCK:	return sizeof (struct nl_namespace_block);
    case NL_ENTRY:		return sizeof (struct nl_entry);
    case NL_CONS:		return sizeof (struct nl_cons);
    case NL_FUNCTION:		return sizeof (struct nl_function);
    default:
      abort ();
    }
}

tree
build_nl_native_function( nl_native_func_t func, int argnum )
{
  tree ret = make_node( NL_NATIVE_FUNCTION );

  NL_NATIVE_FUNCTION_FUNC(ret) = func;
  NL_NATIVE_FUNCTION_ARGNUM(ret) = argnum;

  return ret;
}

tree 
make_nl_entry( tree key, tree value )
{
  tree ret = make_node( NL_ENTRY );
  
  NL_ENTRY_VALUE(ret) = value;
  NL_ENTRY_KEY(ret) = key;

  NL_ENTRY_LAST_DEFINE_LOCATION(ret) = ggc_alloc( sizeof(location_t) );

  return ret;
}
