
#include "c-common.h"

struct lang_identifier GTY(())
{
  struct tree_identifier common;
  enum TOK rid_code;
};

tree build_nl_string( const char *str );
tree copy_nl_string( tree str );

#define NL_STRING_T(NODE) ((struct nl_string*)(NL_STRING_CHECK(NODE)))

#define NL_STRING_LENGTH(NODE) (NL_STRING_T(NODE)->length)
#define NL_STRING_POINTER(NODE) (NL_STRING_T(NODE)->buffer)
#define NL_STRING_SIZE(NODE) (sizeof(struct nl_string) - 1 + NL_STRING_LENGTH(NODE) + 1)

struct nl_string GTY(())
{
  struct tree_common common;
  int length;
  char buffer[1];
};


#define NL_DEF_STMT_T(NODE) ((struct nl_def_stmt*)(NL_DEF_STMT_CHECK (NODE)))
#define NL_DEF_STMT_ID(NODE) (NL_DEF_STMT_T (NODE)->func_ident)
#define NL_DEF_STMT_ARGS(NODE) (NL_DEF_STMT_T (NODE)->arg_ident_list)
#define NL_DEF_STMT_BODY(NODE) (NL_DEF_STMT_T (NODE)->body)

struct nl_def_stmt GTY(())
{
  struct tree_common common;
  tree func_ident;
  tree arg_ident_list;
  tree body;
};


#define ASDF_BLOCK_T(NODE) ((struct asdf_block*)(ASDF_BLOCK_CHECK (NODE)))

#define ASDF_BLOCK_BODY(NODE) (ASDF_BLOCK_T (NODE)->body)
#define ASDF_BLOCK_ENVINFO(NODE) (ASDF_BLOCK_T(NODE).envinfo)

struct asdf_block GTY(())
{
  struct tree_common common;
  tree body;
};

#define NL_LOCAL_VAR_EXPR(NODE) ( (struct nl_local_var*) NL_LOCAL_VAR_EXPR_CHECK(NODE) )

#define NL_LOCAL_VAR_EXPR_STMT_ID(NODE) (NL_LOCAL_VAR_EXPR (NODE)->id)
#define NL_LOCAL_VAR_EXPR_LOCATION(NODE) (NL_LOCAL_VAR_EXPR (NODE)->locus)
#define NL_LOCAL_VAR_EXPR_FILE(NODE) (NL_LOCAL_VAR_EXPR_LOCATION (NODE)->file)
#define NL_LOCAL_VAR_EXPR_LINE(NODE) (NL_LOCAL_VAR_EXPR_LOCATION (NODE)->line)

struct nl_local_var_expr GTY(())
{
  struct tree_common common;
  tree id;
};

#define NL_VAR_STMT_T(NODE) ( (struct nl_var_stmt*)(NL_VAR_STMT_CHECK (NODE)) )

#define NL_VAR_STMT_LOCATION(NODE) ( NL_VAR_STMT_T(NODE)->locus )
#define NL_VAR_STMT_SOURCE_LINE(NODE) ( NL_VAR_STMT_LOCATION(NODE)->line )
#define NL_VAR_STMT_SOURCE_FILE(NODE) ( NL_VAR_STMT_LOCATION(NODE)->file )
#define NL_VAR_STMT_ID(NODE) (NL_VAR_STMT_T (NODE)->id)
#define NL_VAR_STMT_INIT(NODE) (NL_VAR_STMT_T (NODE)->init)

struct nl_var_stmt GTY(())
{
  struct tree_common common;
  location_t *locus;
  tree id;
  tree init;
};

struct nl_special_value GTY(())
{
  struct tree_common common;
};

#define NL_FUNCTION_T(NODE) ( (struct nl_function*)(NL_FUNCTION_CHECK(NODE)) )

#define NL_FUNCTION_ARGS(NODE) ( NL_FUNCTION_T(NODE)->args )
#define NL_FUNCTION_BODY(NODE) ( NL_FUNCTION_T(NODE)->body )
#define NL_FUNCTION_DEFINED_ENV(NODE) ( NL_FUNCTION_T(NODE)->defined_env )
#define NL_FUNCTION_LOCUS(NODE) ( NL_FUNCTION_T(NODE)->locus )
#define NL_FUNCTION_LINE(NODE) ( NL_FUNCTION_LOCUS(NODE)->line )
#define NL_FUNCTION_FILE(NODE) ( NL_FUNCTION_LOCUS(NODE)->file )

struct nl_function GTY(())
{
  struct tree_common common;

  tree args;
  tree body;
  tree defined_env;
  location_t *locus;
};

#define NL_NATIVE_FUNCTION(NODE) ( (struct nl_native_function*)(NL_NATIVE_FUNCTION_CHECK(NODE)) )

#define NL_NATIVE_FUNCTION_FUNC(NODE) ( (nl_native_func_t)(NL_NATIVE_FUNCTION(NODE)->func) )
#define NL_NATIVE_FUNCTION_ARGNUM(NODE) ( NL_NATIVE_FUNCTION(NODE)->argnum )

typedef tree (*nl_native_func_t) (tree);

struct nl_native_function GTY(())
{
  struct tree_common common;
  long func;
  int argnum;
};

tree build_nl_native_function( nl_native_func_t, int argnum );

#define NL_CONS_T(NODE) ( (struct nl_cons*)(NL_CONS_CHECK(NODE)) )

#define NL_CONS_HD(NODE) ( NL_CONS_T(NODE)->hd )
#define NL_CONS_TL(NODE) ( TREE_CHAIN(NL_CONS_T(NODE)) )

struct nl_cons GTY(())
{
  struct tree_common common;
  tree hd;
};

#include "asdf-hash.h"

/* namespace & structure */
struct nl_namespace GTY(())
{
  struct tree_common common;
  struct asdf_hashtab *symtab;
};

#define NL_NAMESPACE_T(NODE) ( (struct nl_namespace*)(NL_NAMESPACE_CHECK(NODE)))
#define NL_NAMESPACE_SYMTAB(NODE) ( NL_NAMESPACE_T(NODE)->symtab )
#define NL_NAMESPACE_CHAIN(NODE) ( TREE_CHAIN(NODE) ) /* for using */
#define NL_NAMESPACE_SUB_NAMESPACE(NODE) ( NL_NAMESPACE_T(NODE)->sub_namespace_list )

/* returns entry */
tree nl_namespace_define( tree ns, tree id, tree value );

tree nl_namespace_lookup( tree ns, tree id, int lookup_chain );

/*
 ** sub_namespace_list
 *
 * namespace x {
 * ...
 * }
 *
 * var y = x;
 *
 * namespace y {
 *  int foo;
 * } 
 *
 *
 * äƤʤäˡx::foo Ƥޤɡ
 *
 */
tree nl_namespace_lookup_entry( tree ns, tree id, int lookup_chain );
tree nl_namespace_create( void );

#define NL_EXPR_STMT_T(NODE) ( (struct nl_expr_stmt*)(NL_EXPR_STMT_CHECK(NODE)) )
#define NL_EXPR_STMT_EXPR(NODE) ( NL_EXPR_STMT_T(NODE)->expr )
#define NL_EXPR_STMT_LOCUS(NODE) ( NL_EXPR_STMT_T(NODE)->locus )
#define NL_EXPR_STMT_SOURCE_LINE(NODE) ( NL_EXPR_STMT_LOCUS(NODE)->line )
#define NL_EXPR_STMT_SOURCE_FILE(NODE) ( NL_EXPR_STMT_LOCUS(NODE)->file )

struct nl_expr_stmt GTY(())
{
  struct tree_common common;
  tree expr;
  location_t *locus;
};

#define NL_NAMESPACE_BLOCK_T(NODE) ( (struct nl_namespace_block*)(NL_NAMESPACE_BLOCK_CHECK(NODE)) )
#define NL_NAMESPACE_BLOCK_BODY(NODE) ( NL_NAMESPACE_BLOCK_T(NODE)->body )
#define NL_NAMESPACE_BLOCK_NAME(NODE) ( NL_NAMESPACE_BLOCK_T(NODE)->name )
#define NL_NAMESPACE_BLOCK_LOCUS(NODE) ( NL_NAMESPACE_BLOCK_T(NODE)->locus )
#define NL_NAMESPACE_BLOCK_SOURCE_LINE(NODE) ( NL_NAMESPACE_BLOCK_LOCUS(NODE)->line )
#define NL_NAMESPACE_BLOCK_SOURCE_FILE(NODE) ( NL_NAMESPACE_BLOCK_LOCUS(NODE)->file )

struct nl_namespace_block GTY(())
{
  struct tree_common common;

  tree name;
  tree body;

  location_t *locus;
};

#define NL_IS_FUNCVAL(NODE) (TREECODE(NODE) == NL_NATIVE_FUNCTION)


struct nl_entry GTY(())
{
  struct tree_common common;

  tree key;
  tree value;

  location_t *last_define;
};

tree make_nl_entry(tree key, tree value);

#define NL_ENTRY_T(NODE) ( (struct nl_entry*)(NL_ENTRY_CHECK(NODE)) )
#define NL_ENTRY_KEY(NODE) ( NL_ENTRY_T(NODE)->key )
#define NL_ENTRY_VALUE(NODE) ( NL_ENTRY_T(NODE)->value )
#define NL_ENTRY_LAST_DEFINE_LOCATION(NODE) ( NL_ENTRY_T(NODE)->last_define )
#define NL_ENTRY_DEFINE_LINE(NODE) ( NL_ENTRY_LAST_DEFINE_LOCATION(NODE)->line )
#define NL_ENTRY_DEFINE_FILE(NODE) ( NL_ENTRY_LAST_DEFINE_LOCATION(NODE)->file )

enum asdf_tree_node_structure_enum {
  TS_ASDF_GENERIC,
  TS_ASDF_IDENTIFIER,
  TS_NL_DEF_STMT,
  TS_ASDF_BLOCK,
  TS_NL_LOCAL_VAR_EXPR,
  TS_NL_VAR_STMT,
  TS_NL_SPECIAL_VALUE,
  TS_NL_NATIVE_FUNCTION,
  TS_NL_NAMESPACE,
  TS_NL_EXPR_STMT,
  TS_NL_NAMESPACE_BLOCK,
  TS_NL_ENTRY,
  TS_NL_CONS,
  TS_NL_STRING,
  TS_NL_FUNCTION,
};

union lang_tree_node 
  GTY((desc ("asdf_tree_node_structure (&%h)"),
       chain_next ("(union lang_tree_node *)TREE_CHAIN (&%h.generic)")))
{
  union tree_node GTY ((tag ("TS_ASDF_GENERIC"), 
			desc ("tree_node_structure (&%h)"))) 
    generic;

  struct nl_def_stmt GTY ((tag ("TS_NL_DEF_STMT"))) nl_def_stmt;
  struct asdf_block GTY ((tag ("TS_ASDF_BLOCK"))) block;
  struct nl_local_var_expr GTY ((tag ("TS_NL_LOCAL_VAR_EXPR"))) lvar_expr;
  struct nl_var_stmt GTY ((tag ("TS_NL_VAR_STMT"))) var_stmt;
  struct nl_special_value GTY ((tag ("TS_NL_SPECIAL_VALUE"))) nl_value;
  struct nl_native_function GTY ((tag ("TS_NL_NATIVE_FUNCTION"))) nl_native_function;
  struct nl_namespace GTY ((tag ("TS_NL_NAMESPACE"))) nl_namespace;
  struct nl_expr_stmt GTY((tag ("TS_NL_EXPR_STMT") )) nl_expr_stmt;
  struct nl_namespace_block GTY((tag ("TS_NL_NAMESPACE_BLOCK"))) nl_namespace_block;
  struct nl_entry GTY((tag ("TS_NL_ENTRY"))) nl_entry;
  struct nl_cons GTY((tag ("TS_NL_CONS") )) nl_cons;
  struct nl_string GTY((tag ("TS_NL_STRING") )) nl_string;
  struct nl_function GTY((tag ("TS_NL_FUNCTION") )) nl_function;
};

enum asdf_tree_node_structure_enum asdf_tree_node_structure(union lang_tree_node *);

struct lang_decl GTY(())
{
};

struct lang_type GTY(())
{
};

struct language_function GTY(())
{
};

typedef struct lang_identifier asdf_identifier;

extern int lineno;

size_t asdf_tree_size (enum tree_code code);

enum QUAL_FLAG {
  QUALIFIER_CONST=1>>0,
  QUALIFIER_STATIC=1>>1,
};

#define DEFTREECODE(SYM,NAME,TYPE,LENGTH) SYM,

enum asdf_tree_code {
  ASDF_DUMMY_TREE_CODE = LAST_C_TREE_CODE,
#include "asdf-tree.def"
  LAST_ASDF_TREE_CODE
};

#undef DEFTREECODE

