#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "toplev.h"

#include "errors.h"

#include "tree.h"
#include "tokens.h"
#include "asdf-tree.h"

#include "compile.h"

static void compile_global_func PARAMS((tree func));
static void compile_global_decls PARAMS((tree program));

void
compile_global_func (func)
     tree func;
{
  tree result_decl;

  input_line = DECL_SOURCE_LINE (func);
  current_function_decl = func;

  announce_function (func);

  make_decl_rtl (func, NULL);

  result_decl = DECL_RESULT (func) = build_decl (RESULT_DECL, NULL_TREE, TREE_TYPE (TREE_TYPE (func)));
  DECL_CONTEXT (result_decl) = func;
  DECL_SOURCE_FILE (result_decl) = input_filename;
  DECL_SOURCE_LINE (result_decl) = input_line;

  DECL_SIZE (result_decl) = 0;

  layout_decl (result_decl, 0);
  DECL_ARGUMENTS (func) = NULL_TREE;

  make_decl_rtl (func, NULL);

  /*
  init_function_start (func);
  expand_function_start (func,0);
  emit_line_note (input_location);
  expand_function_end ( );
  */

  debug_tree (func);

  rest_of_compilation (func);
}

void
compile_global_decls (decls)
     tree decls;
{
  tree decl;

  for ( decl=decls; decl; decl=NL_CONS_TL(decl) ) {
    tree d = NL_CONS_HD(decl);
    switch ( TREE_CODE(d) ) {
    case FUNCTION_DECL:
      compile_global_func (d);
      break;

    case VAR_DECL: {
      tree init = DECL_INITIAL (d);
      input_line = DECL_SOURCE_LINE (d);

      if ( init != 0 ) {
	init = fold (init);

	if ( !TREE_CONSTANT(init) ) {
	  error ("initializer element is not constant");
	  return;
	}
	DECL_INITIAL (d) = fold (init);
      }

      expand_decl (d);
      expand_decl_init (d);
      rest_of_decl_compilation (d, 0, 0, 0);
      break;
    }

    case NL_DEF_STMT:		/* ignore */
    case NL_VAR_STMT:
    case NL_SPECIAL_VALUE:
    case NL_ENTRY:
    case NL_STRING:
    case NL_NATIVE_FUNCTION:
      break;

    case NL_NAMESPACE_BLOCK:	/* fixme*/
      break;

    case NL_EXPR_STMT:		/* fixme */
      break;

    case ASDF_BLOCK:		/* fixme */
      break;

    default:
      internal_error( "compile_global_decls. appears '%s'.", tree_code_name[TREE_CODE(d)] );
      break;
    }
  }
}

void 
compile_program (program)
     tree program;
{
  compile_global_decls (program);
}
