
#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "tree.h"
#include "ggc.h"
#include "decl.h"

#include <stdlib.h>

struct binding_level GTY(())
{
  tree names;
  tree blocks;
  tree cur_block;
  struct binding_level *chain;
};

static GTY (()) struct binding_level *global_level;
static GTY (()) struct binding_level *current_level;

void 
pushlevel(level)
     int level ATTRIBUTE_UNUSED;
{
  struct binding_level *b = ggc_alloc (sizeof(struct binding_level));
  b->chain = current_level;
  b->cur_block = 0;

  current_level = b;
}


tree
poplevel(keep, reverse, functionbody)
     int keep ATTRIBUTE_UNUSED;
     int reverse ATTRIBUTE_UNUSED;
     int functionbody ATTRIBUTE_UNUSED;
{
  struct binding_level *cur = current_level;
  tree block;
  tree decls;
  
  if (reverse)
    cur->names = decls =
      nreverse (current_level->names);
  else
    decls = current_level->names;

  current_level = cur->chain;

  if ( cur->cur_block )
    block = cur->cur_block;
  else
    block = make_node (BLOCK);

  BLOCK_VARS (block) = decls;
  BLOCK_SUBBLOCKS (block) = cur->blocks;

  return block;
}

int 
global_bindings_p(void)
{
  return ( current_level == global_level );
}

void 
insert_block (block)
     tree block ATTRIBUTE_UNUSED;
{
  TREE_USED (block) = 1;
  current_level->blocks = 
    chainon (current_level->blocks, block);
}

void 
set_block (block)
     tree block ATTRIBUTE_UNUSED;
{
  puts( "set_block\n" );
  abort();
}

tree 
pushdecl (decl)
     tree decl;
{
  current_level->names = chainon (current_level->names, decl);
  DECL_CONTEXT (decl) = current_function_decl;

  return decl;
}

tree 
getdecls (void)
{
  return current_level->names;
}

tree 
builtin_function (name, type, function_code, class, library_name, attrs)
     const char *name;
     tree type;
     int function_code;
     enum built_in_class class;
     const char *library_name;
     tree attrs;
{
  fputs("builtin_function",stderr);
  abort();
  
  return NULL_TREE;
}


void
init_global_level()
{
  current_level = global_level = xcalloc( 1, sizeof(*global_level) );
}
