#include "config.h"
#include "system.h"

#include "coretypes-wrapper.h"
#include "input.h"

#include "tree.h"
#include "errors.h"

#include "tokens.h"
#include "lex.h"
#include "nl-eval.h"
#include "asdf-tree.h"
#include "asdf-hash.h"
#include "ggc.h"

#define DEFAULT_SIZE 30

tree
nl_namespace_create( )
{
  tree ret = make_node( NL_NAMESPACE );

  NL_NAMESPACE_SYMTAB(ret) = asdf_hashtab_create(DEFAULT_SIZE);
  NL_NAMESPACE_CHAIN(ret) = NULL;

  return ret;
}

tree
nl_namespace_define( tree ns, tree id, tree value )
{
  tree node = asdf_hashtab_lookup( NL_NAMESPACE_SYMTAB(ns), id, 1 );
  NL_ENTRY_VALUE(node) = value;

  return node;
}

tree
nl_namespace_lookup_entry( tree ns, const tree id, int lookup_chain )
{
  tree p;
  tree node;

  node = asdf_hashtab_lookup( NL_NAMESPACE_SYMTAB(ns), id, 0 );

  if ( (node!=NULL) || 
       (lookup_chain==0) )
    return node;
  
  for ( p=NL_NAMESPACE_CHAIN(ns); p!=NULL; p=NL_NAMESPACE_CHAIN(p) ) {
    tree node = asdf_hashtab_lookup( NL_NAMESPACE_SYMTAB(p), id, 0 );
    
    if ( node != NULL )
      return node;
  }

  return NULL;
}

tree 
nl_namespace_lookup( tree ns, tree id, int chain )
{
  tree entry = nl_namespace_lookup_entry( ns, id, chain );

  if ( entry == NULL )
    return NULL;
  
  return NL_ENTRY_VALUE(entry);
}

tree
nl_lookup_entry( tree id )
{
  tree l;

  for ( l=current_env->namespace_list; l!=NULL; l=TREE_CHAIN(l) ) {
    tree entry = nl_namespace_lookup_entry( TREE_VALUE(l), id, 1 );

    if ( entry != NULL )
      return entry;
  }

  return NULL;
}

tree 
nl_lookup( tree id )
{
  tree entry = nl_lookup_entry( id );

  if ( entry == NULL )
    return NULL;

  return NL_ENTRY_VALUE(entry);
}

tree
nl_define_namespace( tree ns, const char *name )
{
  tree id = get_identifier( name );
  tree ns_p = nl_namespace_lookup( ns, id, 0 );
  tree new_ns;

  if ( ns_p != NULL )
    return ns_p;

  new_ns = nl_namespace_create();

  nl_namespace_define( ns, id, new_ns );

  return new_ns;
}

void
nl_define_native( tree ns, nl_native_func_t func, const char *name, int argnum )
{
  tree id = get_identifier(name);
  tree f = build_nl_native_function( func, argnum );
  nl_namespace_define( ns, id, f );
}

void 
push_namespace( tree ns )
{
  tree cons = tree_cons( NULL, ns, current_env->namespace_list );
  current_env->namespace_list = cons;

  current_namespace = ns;
}

tree 
pop_namespace( void )
{
  tree ret = TREE_VALUE(current_env->namespace_list);
  current_env->namespace_list = TREE_CHAIN(current_env->namespace_list);

  current_namespace = TREE_VALUE( current_env->namespace_list );
  return ret;
}

void
push_environment( tree base_namespace )
{
  struct nl_environment *next = current_env;

  current_env = ggc_alloc( sizeof(struct nl_environment) );
  current_env->chain = next;
  current_env->namespace_list = NULL;

  push_namespace( base_namespace );
}

void
set_environment( tree namespace_list )
{
  struct nl_environment *next = current_env;

  current_env = ggc_alloc( sizeof(struct nl_environment) );
  current_env->namespace_list = namespace_list;
  current_env->chain = next;

  current_namespace = TREE_VALUE( current_env->namespace_list );
}

void
pop_environment( void )
{
  current_env = current_env->chain;

  current_namespace = TREE_VALUE(current_env->namespace_list);
}

