
#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "input.h"
#include "tree.h"
#include "errors.h"

#include "tokens.h"
#include "lex.h"
#include "asdf-tree.h"

#include "nl-eval.h"

static tree special_to_str( tree e );

tree
nl_to_string( tree e )
{
  switch ( TREE_CODE(e) ) {
  case INTEGER_CST: {
    char buf[32];
    sprintf( buf, "%d", (int)TREE_INT_CST_LOW(e) );
    return build_nl_string( buf );
  }

  case NL_STRING:
    return copy_nl_string(e);

  case NL_SPECIAL_VALUE:
    return special_to_str(e);

  default: {
    const char *name = tree_code_name[TREE_CODE(e)];
    tree t= build_nl_string( name );

    return t;
  }
  }
}

static tree 
special_to_str( tree e )
{
  if ( e == nl_true ) 
    return build_nl_string( "true" );
  
  if ( e == nl_false )
    return build_nl_string( "false" );

  if ( e == nl_nil )
    return build_nl_string( "nil" );

  internal_error("special_to_str");
}

static tree
nl_strcat( tree args )
{
  tree arg = args;
  int len = 0;
  int size = 16;
  tree ret;

  char *buf = xcalloc( size+1, 1 );
  buf[0] = '\0';

  while ( arg ) {
    tree str = nl_to_string( NL_CONS_HD(arg) );
    int new_len = len+NL_STRING_LENGTH(str);

    if ( new_len > size ) {
      char *old_buf = buf;

      size = new_len*2;
      buf = xcalloc( size+1, 1 );
      memmove( buf, old_buf, NL_STRING_LENGTH(str) );
    }
    
    strcat( buf, NL_STRING_POINTER(str) );
    len = new_len;

    arg = NL_CONS_TL(arg);
  }

  buf[len] = '\0';

  ret = build_nl_string( buf );

  free( buf );

  return ret;
}
  
static tree
nl_print(tree args)
{
  while ( args != NULL ) {
    tree str = nl_to_string(NL_CONS_HD(args));

    fputs( NL_STRING_POINTER(str), stderr );
    args = NL_CONS_TL(args);
  }
  
  return nl_nil;
}

static tree
nl_gc( tree args )
{
  ggc_collect();
  return nl_nil;
}

void 
nl_init_std(void)
{
  tree nlns = nl_define_namespace( global_namespace, "nl" );
  tree stdns;

  stdns = nl_define_namespace( nlns, "std" );

  nl_define_native( stdns, nl_print, "print", -1 );
  nl_define_native( stdns, nl_to_string, "to_str", 1 );
  nl_define_native( stdns, nl_strcat, "strcat", -1 );
  nl_define_native( stdns, nl_gc, "gc", 0 );
}
