#include "config.h"
#include "system.h"

#include "coretypes-wrapper.h" 
#include "input.h"

#include "tree.h"
#include "errors.h"
#include "toplev.h"

#include "tokens.h"
#include "lex.h"
#include "asdf-tree.h"
#include "parse.h"

#include "compile.h"
#include "decl.h"
#include "nl-eval.h"
#include "ggc.h"

static tree namespace_block PARAMS((int flags));

static tree var_stmt PARAMS((int flags));
static tree for_stmt PARAMS((void));
static tree while_stmt PARAMS((void));
static tree if_stmt PARAMS((void));
static tree statement PARAMS((void));
static tree block PARAMS((void));

static tree expr PARAMS((void));
static tree term_expr PARAMS((void));
static tree scope_expr PARAMS((void));
static tree post_expr PARAMS((void));
static tree mul_expr PARAMS((void));
static tree plus_expr PARAMS((void));
static tree assign_expr PARAMS((void));
static tree compare_expr PARAMS((void));

/*static tree global_decl_var_or_func_or_nlstmt PARAMS((void));
  static tree global_decl_func PARAMS((tree type, tree id));*/
static tree parse_program PARAMS((void));

static tree parse_func_args PARAMS((void));
static tree parse_arg_type PARAMS((void));

static tree nldefun_argdef PARAMS((void));
static tree nldef_stmt PARAMS((void));

/* 
static void unbalance_paren PARAMS((int open_line,int paren_char));

static void
unbalance_paren ( open_line, paren_char )
     int open_line;
     int paren_char;
{
  error ("unbalance '%c'. opens from line %d.", paren_char, open_line);
} 
*/

static int 
parse_qualifier(void)
{
  int flags = 0;

  while ( IS_QUAL_TOKEN(current_token) ) {
    switch ( current_token ) {
    case TOK_CONST:
      flags |= QUALIFIER_CONST;
      break;
      
    default:
      internal_error("parse_qualifier");
      break;
    }

    next_token();
  }

  return flags;
  
}

static void
set_flags( tree t, int flags )
{
  if ( flags & QUALIFIER_CONST )
    TREE_CONSTANT(t) = 1;

  if ( flags & QUALIFIER_STATIC )
    TREE_STATIC(t) = 1;
}

/* fixme: save location */
static tree
if_stmt (void)
{
  tree stmt;
  tree cond, thenbody, elsebody=NULL;

  if ( next_token() != '(' ) {
    error("if statement doesn't have '(' with '%s'.",last_token_str());
    exit (1);
  }

  next_token();
  cond = expr();

  if ( current_token != ')' ) {
    error("if statement lacks ')' with '%s'.", last_token_str());
    exit (1);
  }

  next_token();
  thenbody = statement();

  if ( current_token == TOK_ELSE ) {
    next_token();
    elsebody = statement();
  }

  stmt = make_node( IF_STMT );

  IF_COND(stmt) = cond;
  THEN_CLAUSE(stmt) = thenbody;
  ELSE_CLAUSE(stmt) = elsebody;

  return stmt;
}

static tree
while_stmt(void)
{
  tree body;
  tree cond;
  tree stmt;

  if ( next_token() != '(' ) {
    error("while statement lacks '(' with '%s'", last_token_str());
    exit(1);
  }
  
  next_token();
  cond = expr();

  if ( current_token != ')' ) {
    error("unmatch paren in while statement with '%s'", last_token_str());
    exit(1);
  }

  next_token();
  body = statement();

  stmt = make_node( WHILE_STMT );

  WHILE_COND(stmt) = cond;
  WHILE_BODY(stmt) = body;

  return stmt;
}

static tree
for_stmt(void)
{
  tree init, cond, for_expr, body;
  tree stmt;

  if ( next_token() != '(' ) {
    error("for statement lacks '(' with '%s'", last_token_str());
    exit(1);
  }
  
  next_token();
  init = expr();

  if ( current_token != ';' ) {
    error("for statement must have ';' with '%s'", last_token_str());
    exit(1);
  }

  next_token();
  cond = expr();

  if ( current_token != ';' ) {
    error("for statement must have ';' with '%s'", last_token_str());
    exit(1);
  }

  next_token();
  for_expr = expr();

  if ( current_token != ')' ) {
    error("unmatch paren in for statement with '%s'", last_token_str());
    exit(1);
  }

  next_token();

  body = statement();

  stmt = make_node( FOR_STMT );

  FOR_INIT_STMT(stmt) = init;
  FOR_COND(stmt) = cond;
  FOR_EXPR(stmt) = for_expr;
  FOR_BODY(stmt) = body;    

  return stmt;
}

/*
 * var_stmt ::= 'var' id init_opt;
 * init_opt ::= 
 *          '=' expr
 *        | empty
 */
static tree
var_stmt ( int flags )
{
  tree id;
  tree init;
  tree stmt;

  if ( next_token () != TOK_IDENT ) {
    error ("var statement: %s is not identifier", last_token_str ());
    exit (1);
  }

  id = lex_value.tree_value;
  
  if ( next_token () == '=' ) {
    next_token ();
    init = expr ();
  } else {
    init = NULL_TREE;
  }

  if ( current_token != ';' ) {
    error("var statement lacs ';' with %s.", last_token_str ());
    exit (1);
  }

  stmt = make_node (NL_VAR_STMT);

  NL_VAR_STMT_ID (stmt) = id;
  NL_VAR_STMT_INIT (stmt) = init;

  NL_VAR_STMT_LOCATION (stmt) = ggc_alloc( sizeof(location_t) );

  NL_VAR_STMT_SOURCE_FILE (stmt) = input_filename;
  NL_VAR_STMT_SOURCE_LINE (stmt) = input_line;

  set_flags( stmt, flags );

  next_token();

  return stmt;
}

/*
 * statement ::= 
 *            var_stmt
 *          | decl_stmt
 *          | expr
 */
static tree
statement (void)
{
  tree e;
  int flags = 0;
  int have_qual = 0;

  tree ret;

  if ( IS_QUAL_TOKEN(current_token) ) {
    have_qual = 1;
    flags = parse_qualifier();
  }

  switch ( current_token ) {
  case TOK_VAR:
    return var_stmt( flags );

  case TOK_NAMESPACE:
    return namespace_block( flags );

  case TOK_IF:
    if ( flags != 0 ) 
      error("if statement has qualifiers.");
    return if_stmt( );

  case TOK_WHILE:
    if ( flags != 0 )
      error("while statement has qualifiers.");
    return while_stmt();

  case TOK_FOR:
    if ( flags != 0 )
      error("for statement has qualifiers.");
    return for_stmt();
    
  case TOK_NLDEF:
    if ( flags != 0 )
      error("nldef statement has qualifiers.");
    return nldef_stmt();

  case '{':
    
    return block();

  default: 
    break;
  }

  if ( have_qual ) {
    error ( "expression can't have qualifier." );
  }

  e = expr ();
  ret = make_node( NL_EXPR_STMT );
  NL_EXPR_STMT_EXPR(ret) = e;
  NL_EXPR_STMT_LOCUS(ret) = ggc_alloc( sizeof(location_t ) );
  NL_EXPR_STMT_SOURCE_LINE(ret) = input_line;
  NL_EXPR_STMT_SOURCE_FILE(ret) = input_filename;

  if ( current_token != ';' ) {
    error ("statement lacks ';' with '%s'.", last_token_str () );
    exit (1);
  }

  next_token();

  return ret;
}

static tree
statements( enum TOK last_tok )
{
  tree stmts = NULL;
  tree *s = &stmts;

  while ( current_token != last_tok ) {
    *s = make_node( NL_CONS );
    NL_CONS_HD(*s) = statement();
    s = &(NL_CONS_TL(*s));
  }

  *s = NULL;
  
  return stmts;
}

/*
 * block ::= stmts
 * stmts ::= statement*
 *
 * fixme: to list
 */
static tree
block (void)
{
  tree stmts;
  tree block = make_node( ASDF_BLOCK );

  if ( current_token != '{' ) {
    error ("parse error with '%s'", last_token_str ());
    exit (1);
  }

  next_token();

  stmts = statements( '}' );
  next_token();

  ASDF_BLOCK_BODY( block ) = stmts;
  return block;
}


/*
 * expr :=
 *    epxr = expr
 *  | expr * expr
 *  | expr + expr
 *  | expr / expr
 *  | expr - expr
 *  | ident
 *  | '(' expr ')'
 */

static tree
term_expr (void)
{
  tree val;
  enum TOK tok = current_token;

  switch ( tok ) {
  case TOK_IDENT:
  case TOK_DIGCONST:
  case TOK_STRCONST:
    val = lex_value.tree_value;
    next_token();
    break;

  case TOK_NIL:
    val = nl_nil;
    next_token();
    break;

  case TOK_TRUE:
    val = nl_true;
    next_token();
    break;

  case TOK_FALSE:
    val = nl_false;
    next_token();
    break;
    
  case '(':
    next_token();
    val = expr ();
    break;

  default:
    error ("invalid expressoin with '%s'.", last_token_str ());
    exit (1);
  }

  return val;
}

static tree
scope_expr (void)
{
  tree lval = term_expr();
  tree rval;

  while ( 1 ) {
    if ( current_token != TOK_SCOPE_COLON ) {
      return lval;
    }

    next_token();
    rval = term_expr();
    lval = build( NL_SCOPE_EXPR, NULL, lval, rval );
  }

  /* dummy */
  return lval;

}

static tree
parse_func_args(void)
{
  tree arg;
  tree args;
  tree cons;

  if ( current_token == ')' ) {
    next_token();
    return nl_nil;
  }
  
  arg = expr();
  args = make_node( NL_CONS );
  NL_CONS_HD(args) = arg;
  NL_CONS_TL(args) = nl_nil;

  while ( current_token != ')' ) {
    tree e;
    
    if ( current_token != ',' ) {
      error("funcall arg lacks ',' with '%s'.", last_token_str());
      exit(1);
    }

    next_token();		/* discards ',' */
    e = expr();

    cons = make_node( NL_CONS );
    NL_CONS_HD(cons) = e;
    NL_CONS_TL(cons) = args;

    args = cons;
  }

  next_token();			/* discards ')' */

  return args;
}

static tree
post_expr (void)
{
  tree val = scope_expr();
  
  if ( current_token == '(' ) {
    tree arg, call;
    next_token();
    arg = parse_func_args();
    
    call = build_nt( CALL_EXPR, val, arg, NULL );

    return call;
  }

  return val;
}

static tree
mul_expr (void)
{
  tree lval = post_expr ();
  tree rval;
  enum tree_code op;

  while ( 1 ) {
    if ( current_token == '*' )
      op = MULT_EXPR;
    else if ( current_token == '/' )
      op = TRUNC_DIV_EXPR;
    else 
      return lval;
    
    next_token ();
    rval = post_expr ();
    lval = build (op, integer_type_node, lval, rval);
  }

  /* dummy */
  return lval;
}

static tree
plus_expr (void)
{
  tree lval = mul_expr ();
  tree rval;
  enum tree_code op;

  while ( 1 ) {
    if ( current_token == '+' )
      op = PLUS_EXPR;
    else if ( current_token == '-' )
      op = MINUS_EXPR;
    else 
      return lval;
    next_token ();
    rval = mul_expr ();
    lval = build (op, integer_type_node, lval, rval );
  }
  
  /* dummy */
  return lval;
}

struct tok2op {
  enum TOK tok;
  enum tree_code op;
};

const struct tok2op cmp_tok2op_tbl[] = {
  { TOK_EQUAL,	EQ_EXPR },
  { TOK_NE,	NE_EXPR },
  { '<',	LT_EXPR},
  { TOK_LE,	LE_EXPR},
  { '>',	GT_EXPR},
  { TOK_GE,	GE_EXPR},
};

#define ARY_ELEM(arr) ( sizeof(arr)/sizeof(arr[0]) )

static tree
compare_expr(void)
{
  tree lval = plus_expr();
  tree rval;

  enum tree_code op;

  while ( 1 ) {
    unsigned int i;

    for ( i=0; i<ARY_ELEM(cmp_tok2op_tbl); i++ ) {
      if ( cmp_tok2op_tbl[i].tok == current_token ) {
	op = cmp_tok2op_tbl[i].op;
	break;
      }
    }

    if ( i >= ARY_ELEM(cmp_tok2op_tbl) ) 
      return lval;

    next_token();
    rval = plus_expr();
    lval = build( op, integer_type_node, lval, rval );
  }

  return lval;
}

static tree
assign_expr (void)
{
  tree lval = compare_expr ();
  tree rval;

  tree e;
  
  if ( current_token != '=' ) 
    return lval;

  next_token ();

  rval = assign_expr ();
  e = build ( MODIFY_EXPR, integer_type_node, lval, rval);

  return e;
}

static tree
expr (void)
{
  tree e = assign_expr();

  return e;
}

/*
 * program :=
 *    global_decls
 *
 * global_decls :=
 *     global_decl*
 *
 * global_decl :=
 *    global_decl_var
 *  | global_decl_func
 *  | namespace_block
 *  | type_decl
 *
 * namespace_block :=
 *     global_decls
 */

static tree
namespace_block (int flags)
{
  tree ns;

  enum TOK tok;

  ns = make_node( NL_NAMESPACE_BLOCK );
  NL_NAMESPACE_BLOCK_LOCUS(ns) = ggc_alloc( sizeof(location_t) );

  NL_NAMESPACE_BLOCK_SOURCE_FILE (ns) = input_filename;
  NL_NAMESPACE_BLOCK_SOURCE_LINE (ns) = input_line;

  set_flags( ns, flags );

  tok = next_token();

  if ( tok != TOK_IDENT ) {
    error ("namespace lacks identifier with  '%s'", last_token_str());
    exit( 1 );
  }

  NL_NAMESPACE_BLOCK_NAME(ns) = lex_value.tree_value;

  next_token();			/* current = '{' */
  NL_NAMESPACE_BLOCK_BODY(ns) = block();

  return ns;
}


static tree
parse_arg_type ()
{
  tree arg_type;

  arg_type = tree_cons (NULL_TREE, void_type_node, NULL_TREE );
  arg_type = nreverse( arg_type );

  next_token();			/* { */
  
  return arg_type;
}


#if 0

static tree
global_decl_func(type,id)
     tree type;
     tree id;
{
  tree arg_type = parse_arg_type ();
  tree fn_type = build_function_type (type, arg_type );
  tree fn_decl = build_decl (FUNCTION_DECL, id, fn_type );

  if ( current_token != '{' ) 
    error ("function lacks {");
  else
    next_token ();

  DECL_CONTEXT (fn_decl) = NULL_TREE;
  DECL_SOURCE_FILE (fn_decl) = input_filename;
  DECL_SOURCE_LINE (fn_decl) = input_line;
  TREE_USED (fn_decl) = 1;
  SET_DECL_ASSEMBLER_NAME (fn_decl, DECL_NAME (fn_decl));

  TREE_PUBLIC (fn_decl) = 1;
  DECL_EXTERNAL (fn_decl) = 1; 
  TREE_STATIC (fn_decl) = 0; 

  rest_of_decl_compilation (fn_decl, NULL, 1, 0);
  pushdecl (fn_decl);

  next_token();

  return fn_decl;
}

static tree
global_decl_var_or_func_or_nlstmt()
{
  tree type = expr();
  tree id;
  enum TOK tok = current_token;
  
  if ( tok != TOK_IDENT ) { 

    if ( tok == ';' ) {
      tree stmt = make_node( NL_EXPR_STMT );
      NL_EXPR_STMT_EXPR(stmt) = type;
      NL_EXPR_STMT_LOCUS(stmt) = ggc_alloc( sizeof(location_t) );
      NL_EXPR_STMT_SOURCE_LINE(stmt) = input_line;
      NL_EXPR_STMT_SOURCE_FILE(stmt) = input_filename;

      next_token();		/* discards ';' */

      return stmt;
    }

    error ("declare parse error with '%s'.", last_token_str());
    exit (1);
  }

  id = lex_value.tree_value;
  tok = next_token();

  if ( tok == '(' ) {
    next_token ();
    return global_decl_func( type, id );
  } else {
    /* global_decl_var */
    tree var = build_decl (VAR_DECL, id, type);
    tree init = NULL_TREE;

    if ( tok == '=' ) {
      next_token ();
      init = expr ();
    }

    DECL_INITIAL (var) = init;

    DECL_SOURCE_FILE (var) = input_filename;
    DECL_SOURCE_LINE (var) = input_line;
    DECL_CONTEXT (var) = current_function_decl;

    TREE_PUBLIC (var) = 1;
    TREE_USED (var) = 0;
    TREE_STATIC (var) = 1;
    TREE_TYPE (var) = type;

    DECL_SIZE (var) = 0;

    layout_decl (var, 0);
    
    if ( current_token != ';' ) {
      error ("declaration lacks ';'");
      exit (1);
    } else
      next_token ();

    return var;
  }

  error ("declaration parse error.");
  exit (1);
}
#endif

/*
 * nldef_args_def ::=
 *       '(' nldef_arg1 nldef_args ')'
 *
 * nldef_arg1 ::= identifier
 *
 * nldef_args ::= 
 *          ',' identifier nldef_args
 *        | empty
 */
static tree
nldefun_argdef()
{
  tree args;

  if ( next_token() == ')' ) {
    return nl_nil;
  }

  if ( current_token != TOK_IDENT ) {
    error ("nldef arg error");
    exit (1);
  }

  args = make_node( NL_CONS );
  NL_CONS_HD(args) = lex_value.tree_value;
  NL_CONS_TL(args) = nl_nil;

  while ( next_token () == ',' ) {
    tree cons;

    if ( next_token () != TOK_IDENT ) {
      error ("nldef arg error with '%s'", last_token_str ());
      exit (1);
    }

    cons = make_node( NL_CONS );
    NL_CONS_HD(cons) = lex_value.tree_value;
    NL_CONS_TL(cons) = args;

    args = cons;
  }

  if ( current_token != ')' ) {
    error ("nldef arg error with '%s'", last_token_str ());
    exit (1);
  }
  
  return args;
}

static tree
nldef_stmt()
{
  enum TOK tok = next_token();
  tree defun;

  if ( tok != TOK_IDENT ) {
    error ("'%s' not identifier", last_token_str ());
    exit (1);
  }

  if ( next_token() != '(' ) {
    error ("nldef parse error with '%s'", last_token_str ());
    exit (1);
  }

  defun = make_node( NL_DEF_STMT );
  NL_DEF_STMT_ID (defun) = lex_value.tree_value;
  NL_DEF_STMT_ARGS (defun) = nldefun_argdef();

  next_token();			/* current = '{' */

  NL_DEF_STMT_BODY (defun) = block();

  return defun;
}

/*
static tree
global_decl ()
{
  enum TOK tok = current_token;
  int have_qual=0;
  int flags=0;

  if ( IS_QUAL_TOKEN(tok) ) {
    flags = parse_qualifier();
    have_qual = 0;
  }

  switch ( tok ) {

  case TOK_IDENT:
    return global_decl_var_or_func_or_nlstmt();

  case TOK_NAMESPACE:
    return namespace_block(flags);

  case TOK_NLDEF:
    return nl_defun();

  case TOK_VAR:
    return var_stmt( 0 );

  case TOK_EOF:
  case '}':
    return NULL_TREE;

  default:
    break;
  }

  error ("parse error: not declare '%s'.", last_token_str() );
  exit (1);
}
*/

static tree
parse_program ()
{
  return statements( TOK_EOF );
}
/*
static void test_ggc(void);
static void
test_ggc( void )
{
  int i =0;
  tree val = build_int_2( 255, 0 );
  NL_BEGIN_PROTECT;

  NL_PROTECT(val);

  while ( 1 ) {
    tree x = build_int_2( 100, 0 );
    i++;

    if ( i>100000 ) {
      fprintf( stderr, "p=%d\n", TREE_INT_CST_LOW(val) );
      i = 0;
    }

    ggc_collect();
  }

  NL_END_PROTECT;
}
*/

void 
asdf_parse_file(debug)
     int debug ATTRIBUTE_UNUSED;
{
  tree program;
  NL_BEGIN_PROTECT;

  input_line = 1;

  nl_init();
  next_token();

  program = parse_program();

  NL_PROTECT(program);
  program = nl_eval_program (program);

  /*test_ggc();*/
  compile_program (program);

  NL_END_PROTECT;
}

